import rnaseq_lib.civic
import rnaseq_lib.data
import rnaseq_lib.diff_exp
import rnaseq_lib.dim_red
import rnaseq_lib.docker
import rnaseq_lib.drugs
import rnaseq_lib.graphs
import rnaseq_lib.gtf
import rnaseq_lib.images
import rnaseq_lib.jupyter

import rnaseq_lib.plot
import rnaseq_lib.plot.opts

import rnaseq_lib.tissues

import rnaseq_lib.utils

import rnaseq_lib.web
import rnaseq_lib.web.kegg
import rnaseq_lib.web.openfda
import rnaseq_lib.web.synapse



import os
import shutil
import textwrap
from multiprocessing import cpu_count
from subprocess import call, Popen, PIPE

import numpy as np
import pandas as pd

from rnaseq_lib.docker import fix_directory_ownership, base_docker_call
from rnaseq_lib.tissues import get_tumor_samples, get_gtex_samples, get_normal_samples, map_genes
from rnaseq_lib.utils import mkdir_p


def log2fc(a, b):
    """
    Calculate the log2 Fold Change between two arrays, floats, or integers
    a and b cannot be, nor contain, values less than 0

    :param (int/float/np.array) a: Value or array
    :param (int/float/np.array) b: Value or array
    :return: L2FC array or value
    :rtype: (int/float/np.array)
    """
    return np.log2(a + 1) - np.log2(b + 1)


def run_deseq2(df_path, tissue, output_dir, gtex=True, cores=None):
    """
    Runs DESeq2 on a specific tissue

    :param str df_path: Path to samples by genes dataframe
    :param str tissue: Tissue to run
    :param str output_dir: Full path to output directory
    :param bool gtex: If True uses GTEx as normal tissue. Otherwise uses TCGA Normal
    :param int cores: Number of cores to use. Defaults to # of cores on machine.
    """
    # Make workspace directories
    work_dir = os.path.join(output_dir, 'work_dir')
    mkdir_p(work_dir)

    # Get samples for tissue
    tumor = get_tumor_samples(tissue)
    normal = get_gtex_samples(tissue) if gtex else get_normal_samples(tissue)

    # Write out vectors
    tissue_vector = os.path.join(work_dir, 'tissue.vector')
    with open(tissue_vector, 'w') as f:
        f.write('\n'.join(tumor + normal))

    disease_vector = os.path.join(work_dir, 'disease.vector')
    with open(disease_vector, 'w') as f:
        f.write('\n'.join(['T' if x in tumor else 'N' for x in tumor + normal]))

    # Write out script
    cores = cores if cores else int(cpu_count())
    script_path = os.path.join(work_dir, 'deseq2.R')
    with open(script_path, 'w') as f:
        f.write(
            textwrap.dedent("""
            library('DESeq2'); library('data.table'); library('BiocParallel')
            register(MulticoreParam({cores}))

            # Argument parsing
            args <- commandArgs(trailingOnly = TRUE)
            df_path <- args[1]
            tissue_path <- args[2]
            disease_path <- args[3]
            tissue <- '{tissue}'
            output_dir <- '/data/'

            # Read in vectors
            tissue_vector <- read.table(tissue_path)$V1
            disease_vector <- read.table(disease_path)$V1

            # Read in table and process
            n <- read.table(df_path, sep='\\t', header=1, row.names=1, check.names=FALSE)
            sub <- n[, colnames(n)%in%tissue_vector]
            setcolorder(sub, as.character(tissue_vector))

            # Preprocessing
            countData <- round(sub)
            colData <- data.frame(disease=disease_vector, row.names=colnames(countData))
            y <- DESeqDataSetFromMatrix(countData = countData, colData = colData, design = ~ disease)

            # Run DESeq2
            y <- DESeq(y, parallel=TRUE)
            res <- results(y, parallel=TRUE)
            summary(res)

            # Write out table
            resOrdered <- res[order(res$padj),]
            res_name <- paste(tissue, '.tsv', sep='')
            res_path <- paste(output_dir, res_name, sep='/')
            write.table(as.data.frame(resOrdered), file=res_path, col.names=NA, sep='\\t',  quote=FALSE)

            # MA Plot
            ma_name <- paste(tissue, '-MA.pdf', sep='')
            ma_path <- paste(output_dir, ma_name, sep='/')
            pdf(ma_path, width=7, height=7)
            plotMA(res, main='DESeq2')
            dev.off()

            # Dispersion Plot
            disp_name <- paste(tissue, '-dispersion.pdf', sep='')
            disp_path <- paste(output_dir, disp_name, sep='/')
            pdf(disp_path, width=7, height=7)
            plotDispEsts( y, ylim = c(1e-6, 1e1) )
            dev.off()

            # PVal Hist
            hist_name <- paste(tissue, '-pval-hist.pdf', sep='')
            hist_path <- paste(output_dir, hist_name, sep='/')
            pdf(hist_path, width=7, height=7)
            hist( res$pvalue, breaks=20, col="grey" )
            dev.off()

            # Ratios plots
            qs <- c( 0, quantile( res$baseMean[res$baseMean > 0], 0:7/7 ) )
            bins <- cut( res$baseMean, qs )
            levels(bins) <- paste0("~",round(.5*qs[-1] + .5*qs[-length(qs)]))
            ratios <- tapply( res$pvalue, bins, function(p) mean( p < .01, na.rm=TRUE ) )
            ratio_name <- paste(tissue, '-ratios.pdf', sep='')
            ratio_path <- paste(output_dir, ratio_name, sep='/')
            pdf(ratio_path, width=7, height=7)
            barplot(ratios, xlab="mean normalized count", ylab="ratio of small $p$ values")
            dev.off()                                           
            """.format(cores=cores, tissue=tissue)))

    # Call DESeq2
    docker_parameters = ['docker', 'run',
                         '-v', '{}:/data'.format(output_dir),
                         '-v', '{}:/df'.format(os.path.dirname(df_path)),
                         'jvivian/deseq2']

    parameters = ['/data/work_dir/deseq2.R',
                  '/df/{}'.format(os.path.basename(df_path)),
                  '/data/{}'.format(os.path.join('work_dir', 'tissue.vector')),
                  '/data/{}'.format(os.path.join('work_dir', 'disease.vector'))]

    print '\nCalling: {}\n'.format(' '.join(docker_parameters + parameters))
    p = Popen(docker_parameters + parameters, stderr=PIPE, stdout=PIPE)
    out, err = p.communicate()
    if out or err:
        print out
        print err

    # Fix output of files
    fix_directory_ownership(output_dir=output_dir, tool='jvivian/deseq2')

    # Add gene names to output
    output_tsv = os.path.join(output_dir, '{}.tsv'.format(tissue))
    df = pd.read_csv(output_tsv, index_col=0, sep='\t')
    df.index = map_genes(df.index)
    df.to_csv(output_tsv, sep='\t')

    # Clean up
    shutil.rmtree(work_dir)


def deseq2_normalize(df_path,
                     output_dir='.',
                     map_gene_names=True,
                     clean_workdir=True,
                     normalize_fn=None,
                     suffix='deseq2-normalized.tsv'):
    """
    Accepts a gene by sample expression matrix normalized values with DESeq2
    Output filename: <INPUT>.deseq2-normalized.tsv

    :param str df_path: Path to input expression gene by sample dataframe
    :param str output_dir: Output directory
    :param bool map_gene_names: If True, maps gene IDs to gene names
    :param bool clean_workdir: If True, deletes temporary work directory
    :param fn normalize_fn: Pass a function to apply to the dataframe before normalization. e.g. lambda x: 2**x + 1
    :param str suffix: Suffix added after df_paths basename
    :return: Path to normalized dataframe
    :rtype: str
    """
    # Make workspace directory
    output_dir = os.path.abspath(output_dir)
    work_dir = os.path.join(output_dir, 'work_dir')
    mkdir_p(work_dir)

    # Write out pseudo-vector
    # Tested that results are the same regardless of vector grouping
    samples = [x.strip() for x in open(df_path, 'r').readline().split()[1:]]
    tissue_vector = os.path.join(work_dir, 'tissue.vector')
    with open(tissue_vector, 'w') as f:
        f.write('\n'.join(samples))

    if normalize_fn:
        df = pd.read_csv(df_path, sep='\t', index_col=0)
        df = df.apply(normalize_fn)
        df_path = os.path.join(os.path.dirname(df_path), 'processed.' + os.path.basename(df_path))
        df.to_csv(df_path, sep='\t')

    # Write normalization script
    output_path = os.path.join(output_dir, os.path.basename(df_path).split('.')[0] + suffix)
    script_path = os.path.join(work_dir, 'deseq2.R')
    with open(script_path, 'w') as f:
        f.write(
            textwrap.dedent("""
            suppressMessages(library('DESeq2'))

            # Argument parsing
            output_dir <- '/data/'

            # Read in pseudo-vector
            tissue_vector <- read.table('/data/work_dir/tissue.vector')$V1

            # Read in table and process
            print("Reading in dataframe")
            n <- read.table('/df/{df_path}', sep='\\t', header=1, row.names=1, check.names=FALSE)

            # Preprocessing
            print("Rounding data to integers")
            countData <- round(n)

            print("Creating DESeq2 Dataset Object")
            colData <- data.frame(tissue=tissue_vector, row.names=colnames(countData))
            dds <- DESeqDataSetFromMatrix(countData = countData, colData = colData, design = ~ tissue)

            # Estimate size factors
            print("Estimating Size Factors")
            dds <- estimateSizeFactors(dds)

            # Extract Normalized Counts and Write
            print("Extracting normalized counts")
            norm <- counts(dds, normalized = TRUE)

            print("Writing output: {output_name}")
            write.table(norm, file="/data/{output_name}", sep='\\t', quote=F, dec='.', col.names=NA) 
            """.format(df_path=os.path.basename(df_path), output_name=os.path.basename(output_path))))

    # Call Docker
    base_params = base_docker_call(os.path.dirname(output_path))
    parameters = base_params + ['-v', '{}:/df'.format(os.path.abspath(os.path.dirname(df_path))),
                                'jvivian/deseq2',
                                '/data/work_dir/deseq2.R']

    print '\nCalling: {}\n'.format(' '.join(parameters))
    call(parameters)

    # Fix output of files
    fix_directory_ownership(output_dir=output_dir, tool='jvivian/deseq2')

    # Map gene IDs to gene names
    if map_gene_names:
        df = pd.read_csv(output_path, index_col=0, sep='\t')
        df.index = map_genes(df.index)
        df.to_csv(output_path, sep='\t')

    # Clean up
    if clean_workdir:
        shutil.rmtree(work_dir)

    return output_path



import os
from itertools import combinations

import pandas as pd


def output_gephi(df, node_attrs, output_dir):
    """
    Output two CSVs used by Gephi to create graphs:
        nodes.csv
        edges.csv

    :param pd.DataFrame df: Dataframe of input data
    :param list(str) node_attrs: List of attributes in DF to create nodes from
    :param str output_dir: Path to output directory
    """
    # Create nodes
    nodes = []
    for attr in node_attrs:
        nodes.extend(map(lambda x: (x, attr), df[attr].tolist()))
    nodes = pd.DataFrame.from_records(nodes, columns=('Label', 'Type'))

    # Create edges
    edges = set()
    for row in df.iterrows():
        i, row = row
        for j, k in combinations(node_attrs, 2):
            edges.add((int(nodes[nodes.Label == row[j]].index[0]), int(nodes[nodes.Label == row[k]].index[0])))
    edges = pd.DataFrame.from_records(list(edges), columns=('Source', 'Target'))

    # Output
    nodes.to_csv(os.path.join(output_dir, 'nodes.csv'), index=True, index_label='Id')
    edges.to_csv(os.path.join(output_dir, 'edges.csv'), index=False)



#!/usr/bin/python -tt
# -*- coding: utf-8 -*-
"""
Created on Fri May 26 12:17:14 2017

@author: ehsanamid
"""

import sys
import time

import numpy as np
# from sklearn.decomposition import PCA
from sklearn.decomposition import TruncatedSVD
from sklearn.neighbors import NearestNeighbors as knn


def generate_triplets(X, kin=50, kout=10, kr=5, weight_adj=False, random_triplets=True):
    num_extra = np.maximum(kin + 50, 50)  # look up more neighbors
    #    num_extra = kin
    #    kout = 20
    n = X.shape[0]
    nbrs = knn(n_neighbors=num_extra + 1, algorithm='auto').fit(X)
    distances, indices = nbrs.kneighbors(X)
    #    sig = distances[:,10]
    sig = np.maximum(np.mean(distances[:, 10:20], axis=1), 1e-20)  # scale parameter
    P = np.exp(-distances ** 2 / np.reshape(sig[indices.flatten()], [n, num_extra + 1]) / sig[:, np.newaxis])
    sort_indices = np.argsort(-P, axis=1)  # actual neighbors

    triplets = np.zeros([n * kin * kout, 3])
    weights = np.zeros(n * kin * kout)

    cnt = 0
    for i in xrange(n):
        for j in xrange(kin):
            sim = indices[i, sort_indices[i, j + 1]]
            p_sim = P[i, sort_indices[i, j + 1]]
            rem = indices[i, sort_indices[i, :j + 2]].tolist()
            l = 0
            while (l < kout):
                out = np.random.choice(n)
                if out not in rem:
                    triplets[cnt, :] = [i, sim, out]
                    p_out = np.exp(-np.sum((X[i, :] - X[out, :]) ** 2) / (sig[i] * sig[out]))
                    if p_out < 1e-20:
                        p_out = 1e-20
                    weights[cnt] = p_sim / p_out
                    rem.append(out)
                    l += 1
                    cnt += 1
        if ((i + 1) % 10000) == 0:
            print 'Genareted triplets %d out of %d' % (i + 1, n)
    if random_triplets:
        kr = 5
        triplets_rand = np.zeros([n * kr, 3])
        weights_rand = np.zeros(n * kr)
        for i in xrange(n):
            cnt = 0
            while cnt < kr:
                sim = np.random.choice(n)
                if sim == i:
                    continue
                out = np.random.choice(n)
                if out == i or out == sim:
                    continue
                p_sim = np.exp(-np.sum((X[i, :] - X[sim, :]) ** 2) / (sig[i] * sig[sim]))
                if p_sim < 1e-20:
                    p_sim = 1e-20
                p_out = np.exp(-np.sum((X[i, :] - X[out, :]) ** 2) / (sig[i] * sig[out]))
                if p_out < 1e-20:
                    p_out = 1e-20
                if p_sim < p_out:
                    sim, out = out, sim
                    p_sim, p_out = p_out, p_sim
                triplets_rand[i * kr + cnt, :] = [i, sim, out]
                weights_rand[i * kr + cnt] = p_sim / p_out
                cnt += 1
            if ((i + 1) % 10000) == 0:
                print 'Genareted random triplets %d out of %d' % (i + 1, n)
        triplets = np.vstack((triplets, triplets_rand))
        weights = np.hstack((weights, weights_rand))
    triplets = triplets[~np.isnan(weights), :]
    weights = weights[~np.isnan(weights)]
    weights /= np.max(weights)
    weights += 0.001
    if weight_adj:
        weights = np.log(1 + 50 * weights)
        weights /= np.max(weights)
    return (triplets.astype(int), weights.flatten())


def tete_grad(Y, triplets, weights):
    n, dim = Y.shape
    grad = np.zeros([n, dim])
    y_ij = Y[triplets[:, 0], :] - Y[triplets[:, 1], :]
    y_ik = Y[triplets[:, 0], :] - Y[triplets[:, 2], :]
    d_ij = 1 + np.sum(y_ij ** 2, axis=-1)
    d_ik = 1 + np.sum(y_ik ** 2, axis=-1)
    num_viol = np.sum(d_ij > d_ik)
    denom = (d_ij + d_ik) ** 2
    loss = weights.dot(d_ij / (d_ij + d_ik))
    gs = 2 * y_ij * (d_ik / denom * weights)[:, np.newaxis]
    go = 2 * y_ik * (d_ij / denom * weights)[:, np.newaxis]
    for i in range(dim):
        grad[:, i] += np.bincount(triplets[:, 0], weights=gs[:, i] - go[:, i])
        grad[:, i] += np.bincount(triplets[:, 1], weights=-gs[:, i])
        grad[:, i] += np.bincount(triplets[:, 2], weights=go[:, i])
    return (loss, grad, num_viol)


def tete(X, num_dims=2, num_neighbs=50, num_out=10, num_rand=5, eta=1000.0, Yinit=[]):
    n, dim = X.shape
    X -= np.min(X)
    X /= np.max(X)
    X -= np.mean(X, axis=0)
    if dim > 50:
        #        pca = PCA(n_components=50)
        #        pca.fit(X)
        #        X = np.dot(X, pca.components_.transpose())
        X = TruncatedSVD(n_components=50, random_state=0).fit_transform(X)
    if np.size(Yinit) > 0:
        Y = Yinit
    else:
        Y = np.random.normal(size=[n, num_dims]) * 0.0001
    C = np.inf
    best_C = np.inf
    best_Y = Y
    tol = 1e-7
    num_iter = 2000
    #    eta = 500.0 # learning rate

    triplets, weights = generate_triplets(X, num_neighbs, num_out, num_rand)
    num_triplets = float(triplets.shape[0])

    for itr in range(num_iter):
        old_C = C
        C, grad, num_viol = tete_grad(Y, triplets, weights)

        # maintain the best answer
        if C < best_C:
            best_C = C
            best_Y = Y

        # update Y
        Y -= (eta / num_triplets * n) * grad;

        # update the learning rate
        if old_C > C + tol:
            eta = eta * 1.01
        else:
            eta = eta * 0.5

        if (itr + 1) % 100 == 0:
            print 'Iteration: %4d, Loss: %3.3f, Violated triplets: %0.4f' % (itr + 1, C, float(num_viol) / num_triplets)
    return best_Y


def load_known_size(fname):
    t = time.time()
    first = True
    with open(fname) as f:
        for irow, line in enumerate(f):
            if first:
                nrow, ncol = [int(a) for a in line.split()]
                x = np.empty((nrow, ncol), dtype=np.double)
                first = False
            else:
                x[irow - 1, :] = [float(a) for a in line.split()]
    print "Elapsed time %s" % (time.time() - t)
    return x


def main():
    filename = sys.argv[1]
    X = load_known_size(filename)
    if len(sys.argv) > 2:
        init = sys.argv[2]
        Yinit = np.loadtxt(init)
        Y = tete(X, 2, 50, Yinit)
    else:
        Yinit = np.random.normal(size=[X.shape[0], 2]) * 0.0001
        Y = tete(X, 2, 50, Yinit)
        # np.savetxt('../data/' + filename[filename.find('/',3)+1:filename.find('.',3)] + '_init_sb.txt',Yinit)
    np.savetxt('../results/' + filename[filename.find('/', 3) + 1:filename.find('.', 3)] + '_2D_sb.txt', Y)


if __name__ == '__main__':
    main()



import io
import os

import pandas as pd
from Bio.KEGG.KGML import KGML_parser

from rnaseq_lib.web import _rget


# Gene Functions
def find_gene_label(query, database='genes', human_only=True):
    r = _kegg_search(operation='find', database=database, query=query)
    if r:
        for hit in r.text.split('\n'):
            hit = hit.split('\t')[0]
            if human_only:
                if hit.startswith('hsa'):
                    return str(hit)
            else:
                return str(hit)
    else:
        return None


def get_gene_dataframe(gene):
    label = find_gene_label(gene)
    r = _get(label)
    return _parse_gene_get(r.text)


def get_gene_names_from_label(label):
    genes = set()
    r = _get(label)
    for line in r.text.split('\n'):
        if line.startswith('NAME'):
            line = line.split()[1:]
            genes.add(line[0].rstrip(','))
    return genes


# Drug Functions
def find_drug_label(drug):
    r = _kegg_search(operation='find', database='drug', query=drug)
    for line in r.text.split('\n'):
        try:
            return line.split()[0]
        except IndexError:
            print 'No entry for {}, check spelling'.format(drug)


def get_drug_info(drug):
    label = find_drug_label(drug)
    r = _get(label)
    return r.text


def get_drug_class(drug):
    text = get_drug_info(drug)
    values = []

    # Flag for once the BRITE is reached
    brite = False
    for line in text.split('\n'):
        line = line.strip().split()
        if not line:
            break
        if line[0] == 'BRITE':
            brite = True
        if line[0].startswith('L') and brite:
            values.append(' '.join(line))

    # Handle returning values (or notify user if no values found)
    if text and not values:
        print 'Response found for {}, but no drug class provided'.format(drug)
    else:
        return values


def get_drug_disease(drug):
    text = get_drug_info(drug)
    values = []

    disease = True
    for line in text.split('\n'):
        line = line.strip().split()
        if not line:
            break
        if line[0].isupper() and disease:
            break
        if line[0].upper() == 'DISEASE':
            disease = True
        if disease:
            line.remove('DISEASE')
            values.append(' '.join(line))


# Pathway Functions
def find_pathway(query):
    return _kegg_search(operation='find', database='pathway', query=query)


def get_kgml(pathway):
    kgml = _get(pathway, form='kgml').text

    # Wrap text in a file handle for KGML parser
    f = io.BytesIO(kgml.encode('utf-8'))
    k = KGML_parser.read(f)
    return k


def get_pathway_info(pathway):
    return _get(pathway).text


def pathway_image(pathway):
    from IPython.display import Image
    k = get_kgml(pathway)
    return Image(k.image)


def get_genes_from_pathway(pathway):
    kgml = _get(pathway, form='kgml').text

    # Wrap text in a file handle for KGML parser
    f = io.BytesIO(kgml.encode('utf-8'))
    k = KGML_parser.read(f)

    genes = set()
    for gene in k.genes:
        g = get_gene_names_from_label(gene)
        genes = genes.union(g)

    return genes


# Internal Functions
def _kegg_search(operation, database, query=None, form=None):
    # Set arguments to empty strings if None
    query = '' if query is None else query
    form = '' if form is None else form

    # Define base URL
    url = 'http://rest.kegg.jp'

    # Make get request
    request = os.path.join(url, operation, database, query, form)
    return _rget(request)


def _get(query, form=None):
    return _kegg_search(operation='get', database='', query=query, form=form)


def _parse_drug_line(line, index=1):
    drug_value = None
    for item in line[index:]:
        if item.startswith('D') and not item.startswith('DG'):
            drug_value = item
            break
    return drug_value


def _parse_gene_get(text):
    # Entries in the text we want to include
    entries = ['ORTHOLOGY', 'ORGANISM', 'PATHWAY', 'NETWORK', 'DISEASE']

    # Create first entry (header) in dataframe
    records = [('Category', 'Kegg', 'Entry')]

    # Flag variable for specifying current entry
    current_entry = None

    for line in text.split('\n'):

        # If line is empty or contains trailing slashes from request
        if line.isspace() or line.startswith('/'):
            break

        # Split line by whitespace as there is no standard delimiter
        line = line.split()

        # First line
        if line[0] == 'ENTRY':
            pass

        # Name row has no Kegg value
        elif line[0] == 'NAME':
            records.append(('Name', None, ' '.join(line[1:])))

        # Definition also has no Kegg value
        elif line[0] == 'DEFINITION':
            records.append(('Definition', None, ' '.join(line[1:])))

        # Drug target row is non-standard, requires separate parsing
        elif line[0] == 'DRUG_TARGET':
            drug_value = _parse_drug_line(line, index=2)
            records.append(('Drug', drug_value, line[1]))
            current_entry = 'Drug'

        # Continuation of drug entries
        elif current_entry == 'Drug':
            drug_value = _parse_drug_line(line, index=1)
            records.append(('Drug', drug_value, line[0]))

        # First line for standard entries
        elif line[0] in entries:
            name = line[0].lower().capitalize()
            records.append((name, line[1], ' '.join(line[1:])))
            current_entry = name

        # Stopping condition -- the rest of the table is hard to parse and less interesting
        elif line[0] == 'BRITE':
            break

        # If in an entry and not yet at the stopping condition
        elif current_entry:
            records.append((current_entry, line[0], ' '.join(line[1:])))

    # Return pandas dataframe
    return pd.DataFrame.from_records(records)



import os
import pickle

import pandas as pd

_cwd = os.path.realpath(os.path.join(os.getcwd(), os.path.dirname(__file__)))


def load_samples():
    """
    Returns sample dictionary which maps TCGA and GTEx samples to a tissue.
    Synapse ID: syn10296681

    :return: Tissues are keys are list of samples are values
    :rtype: dict(str, list(str))
    """
    return pickle.load(open(os.path.join(_cwd, 'samples.pickle'), 'rb'))


def load_gene_map():
    """
    Dictionary mapping gene ID to gene name

    :return: Gene map
    :rtype: dict
    """
    return pickle.load(open(os.path.join(_cwd, 'gene_map.pickle'), 'rb'))


def load_mab_targets():
    """
    Returns sorted list of MAB cancer drug targets

    :return: Sorted gene list
    :rtype: list
    """
    path = os.path.join(_cwd, 'cancer-MAB-gene-targets.txt')
    return sorted([x.strip() for x in open(path, 'r').readlines()])


def load_drug_gene_tissue_table():
    """
    Loads Dataframe containing Drug, Gene, Tissue, and DE information

    :return: Dataframe
    :rtype: pd.DataFrame
    """
    return pd.read_csv(os.path.join(_cwd, 'drugs', 'drug-gene-tissue-deseq2.tsv'), sep='\t', index_col=0)


def load_gene_tissue_drug_map():
    """
    Tissue and Drug information for a given gene

    :return: Mapping
    :rtype: dict(str, list(str))
    """
    df = load_drug_gene_tissue_table()
    return {k: {'tissue': set(g['tissue'].tolist()), 'drug': set(g['generic_name'].tolist())}
            for k, g in df.groupby('gene')}


def load_ucsf_genes():
    """
    Returns sorted list of UCSF genes

    :return: Sorted gene list
    :rtype: list
    """
    path = os.path.join(_cwd, 'UCSF-genes.csv')
    return sorted([x.strip() for x in open(path, 'r').readlines()])


def load_civic_genes():
    """
    Returns sorted list of genes from CIViC

    :return: Sorted gene list
    :rtype: list
    """
    path = os.path.join(_cwd, 'data/civic-genes.txt')
    return sorted([x.strip() for x in open(path, 'r').readlines()])



import errno
import os

from rnaseq_lib.utils.expando import Expando

_iter_types = (list, tuple, set, frozenset)


def mkdir_p(path):
    """
    Creates directory unless it already exists

    :param str path: Path of directory to make
    """
    try:
        os.makedirs(path)
    except OSError as exc:
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise


def flatten(x):
    """
    Flattens a nested array into a single list

    :param list x: The nested list/tuple to be flattened
    """
    result = []
    for el in x:
        if hasattr(el, "__iter__") and not isinstance(el, basestring):
            result.extend(flatten(el))
        else:
            result.append(el)
    return result


def merge_dicts(x, y):
    """
    Given two dicts, merge them into a new dict as a shallow copy

    param dict x: first dictionary
    param dict y: second dictionary
    """
    z = x.copy()
    z.update(y)
    return z


def which(name, path=None):
    """
    Look for an executable file of the given name in the given list of directories,
    or the directories listed in the PATH variable of the current environment. Roughly the
    equivalent of the `which` program. Does not work on Windows.

    :type name: str
    :param name: the name of the program

    :type path: Iterable
    :param path: the directory paths to consider or None if the directories referenced in the
    PATH environment variable should be used instead

    :returns: an iterator yielding the full path to every occurrance of an executable file of the
    given name in a directory on the given path or the PATH environment variable if no path was
    passed

    >>> next( which('ls') )
    '/bin/ls'
    >>> list( which('asdalskhvxjvkjhsdasdnbmfiewwewe') )
    []
    >>> list( which('ls', path=()) )
    []
    """
    if path is None:
        path = os.environ.get('PATH')
        if path is None:
            return
        path = path.split(os.pathsep)
    for bin_dir in path:
        executable_path = os.path.join(bin_dir, name)
        if os.access(executable_path, os.X_OK):
            yield executable_path


def rexpando(d):
    """
    Recursive Expando!

    Recursively iterate through a nested dict / list object
    to convert all dictionaries to Expando objects

    :param dict d: Dictionary to convert to nested Expando objects
    :return: Converted dictionary
    :rtype: Expando
    """
    e = Expando()
    for k, v in d.iteritems():
        k = _key_to_attribute(k)
        if isinstance(v, dict):
            e[k] = rexpando(v)
        elif isinstance(v, _iter_types):
            e[k] = _rexpando_iter_helper(v)
        else:
            e[k] = v
    return e


def _rexpando_iter_helper(input_iter):
    """
    Recursively handle iterables for rexpando

    :param iter input_iter: Iterable to process
    :return: Processed iterable
    :rtype: list
    """
    l = []
    for v in input_iter:
        if isinstance(v, dict):
            l.append(rexpando(v))
        elif isinstance(v, _iter_types):
            l.append(_rexpando_iter_helper(v))
        else:
            l.append(v)
    return l


def _key_to_attribute(key):
    """
    Processes key for attribute accession by replacing illegal chars with a single '_'

    :param str key: Dictionary key to process
    :return: Processed key
    :rtype: str
    """
    illegal_chars = ['!', '@', '#', '$', '%', '^', '&', '*', '(', ')', '-', '.', ',', '+', '/', '\\', ':', ';']
    for c in illegal_chars:
        key = key.replace(c, '_')
    return '_'.join(x for x in key.split('_') if x)  # Remove superfluous '_' chars



import os

import pandas as pd

__location__ = os.path.dirname(os.path.realpath(os.path.join(os.getcwd(), os.path.dirname(__file__))))


def get_anova():
    return pd.read_excel(os.path.join(__location__, 'data/cancerrxgene/ANOVA_OUTPUT.xlsx'))


def get_cell_line():
    return pd.read_excel(os.path.join(__location__, 'data/cancerrxgene/Cell_Lines_Details.xlsx'))


def get_drugs():
    return pd.read_excel(os.path.join(__location__, 'data/cancerrxgene/Screened_Compounds.xlsx'))


def get_dose_response():
    return pd.read_excel(os.path.join(__location__, 'data/cancerrxgene/Screened_Compounds.xlsx'))



import time
from collections import defaultdict

import pandas as pd
from progressbar import ProgressBar
from rnaseq_lib.utils import rexpando
from rnaseq_lib.web import _rget
from rnaseq_lib.tissues import grep_cancer_terms


def query_drug(drug):
    """
    Search OpenFDA drug label API by name

    :param str drug: Drug to query by
    :return: Response containing first match
    :rtype: requests.models.Response
    """
    # Clean input
    drug = drug.lower()

    # Look for drug label on openFDA
    for d in [drug, drug.capitalize(), drug.upper()]:
        for name in ['generic_name', 'brand_name']:
            url = 'https://api.fda.gov/drug/label.json?search=openfda.{}:{}&limit=1'.format(name, d)
            time.sleep(0.1)
            r = _rget(url)
            if r:
                return r
    return None


def drugs_by_query(query, limit=100, field='indications_and_usage'):
    """
    Search OpenFDA API for drugs that match query

    :param str query: Query string to search by
    :param int limit: Limit request, cannot be higher than 100
    :param str field: OpenFDA field to search. Example: "openfda.brand_name"
    :return: Return drugs by query term
    :rtype: list(tuple(str, str))
    """
    assert limit <= 100, 'OpenFDA API does not allow a limit higher than 100'
    url = 'https://api.fda.gov/drug/label.json?search={}:{}&limit={}'.format(field, query, limit)
    r = _rget(url)
    if r:
        # Convert to JSON
        r = r.json()

        # Get total number of hits
        total_terms = r['meta']['results']['total']
        print 'Found a total of {} terms'.format(total_terms)

        # Collect first batch
        drugs = [(','.join(x['openfda']['brand_name']), ','.join(x['openfda']['generic_name']))
                 for x in r['results']]

        # Collect results in batches
        bar = ProgressBar()
        skip = limit
        for _ in bar(xrange(int(total_terms) / limit)):
            time.sleep(1)
            print 'Collecting samples {} - {}'.format(skip, skip + limit)
            r = _rget(url + '&skip={}'.format(skip))
            if r:
                r = r.json()
                drugs.extend([(','.join(x['openfda']['brand_name']), ','.join(x['openfda']['generic_name']))
                              for x in r['results']])
                skip += limit
        return drugs
    else:
        return None


def drugs_to_dataframe(drugs):
    """
    Convert a list of drugs to an OpenFDA DataFrame

    :param list(str) drugs:
    :return: DataFrame of OpenFDA information
    """
    # Create dictionary to store table info
    info = defaultdict(list)

    # For each drug, query openFDA
    bar = ProgressBar()
    for drug in bar(drugs):
        drug = drug.lower()
        r = query_drug(drug)
        if r:
            hits = rexpando(r.json()).results

            # If more than one hit is returned, find exact match
            if len(hits) != 1:
                for h in hits:
                    if drug in h.openfda.brand_name.lower() or drug in h.openfda.generic_name.lower():
                        res = h
            else:
                res = hits[0]

            # Collect info if description references cancer
            if 'indications_and_usage' in res:
                usage = res.indications_and_usage[0] if type(
                    res.indications_and_usage) is list else res.indications_and_usage
                if [x for x in usage.replace('\n', '.').split('.') if 'cancer' in x]:

                    # Get generic name
                    try:
                        info['generic_name'].append(str(res.openfda.generic_name).strip("u'[]"))
                    except AttributeError:
                        info['generic_name'].append(None)

                    # Get brand name
                    try:
                        info['brand_name'].append(str(res.openfda.brand_name).strip("u'[]"))
                    except AttributeError:
                        info['brand_name'].append(None)

                    # Get usage
                    try:
                        info['usage'].append(str(res.indications_and_usage).strip('[]'))
                    except AttributeError:
                        info['usage'].append(None)

                    # Get mechanism of action
                    try:
                        info['mech_action'].append(str(res.mechanism_of_action).strip('[]'))
                    except AttributeError:
                        try:
                            info['mech_action'].append(str(res.clinical_pharmacology).strip('[]'))
                        except AttributeError:
                            info['mech_action'].append(None)

    return pd.DataFrame.from_dict(info)


def indications_and_usage(drug):
    """
    Returns indications and usage, split by cancer terms

    :param str drug: Name of drug
    :return: Indications and usage split by cancer terms
    :rtype: str
    """
    r = query_drug(drug)
    if not r:
        print 'No drug found'
        return None

    return '\n\n'.join(grep_cancer_terms(str(rexpando(
        r.json()['results'][0])['indications_and_usage'])))


import holoviews as hv
import numpy as np
import pandas as pd

from rnaseq_lib.diff_exp import log2fc
from rnaseq_lib.dim_red import run_tsne, run_tete
from rnaseq_lib.plot.opts import gene_curves_opts, gene_kde_opts, gene_distribution_opts, gene_de_opts
from rnaseq_lib.tissues import subset_by_dataset


class Holoview:
    """
    Object for Holoviews plots of gene expression data. Created for use with Holomap and DynamicMap which cannot
    accept dataframes as arguments. This class circumvents that limitation by referencing
    the dataframe internally.
    """

    def __init__(self, df):
        """
        :param pd.DataFrame df: Dataframe containing metadata / expression values (Synapse.org: syn11515015)
        """
        self.df = df
        self.df_cols = ['id', 'tissue', 'dataset', 'tumor', 'type']

        # Style attributes - used in conjunction with '.opts()'
        self.gene_curves_opts = gene_curves_opts
        self.gene_kde_opts = gene_kde_opts
        self.gene_distribution_opts = gene_distribution_opts
        self.gene_de_opts = gene_de_opts

    @staticmethod
    def l2norm(x):
        """
        Log2 noramlization function for gene counts

        :param float x: Input value
        :return: log2(x+1) normalized value
        :rtype: float
        """
        return np.log2(x + 1)

    def _subset(self, gene, tissue):
        """
        Subset dataframe by gene and tissue with default columns `self.df_cols`

        :param str gene: Gene (ex: ERBB2) to select
        :param str tissue: Tissue (ex: Breast) to select
        :return: Subset dataframe
        :rtype: pd.DataFrame
        """
        df = self.df[self.df_cols + [gene]].sort_values(gene, ascending=False)
        return df[df.tissue == tissue]

    def _gene_cutoff(self, gene, tissue, percent):
        # Subset dataframe by tissue and gene
        df = self._subset(gene, tissue)

        # Subset by dataset
        tumor, normal, gtex = subset_by_dataset(df)

        # Calculate gene expression cutoffs for each dataset
        cutoffs = [x[gene].apply(self.l2norm).sort_values(ascending=False).iloc[int(len(x) * percent) - 1]
                   for x in [tumor, normal, gtex]]

        # Return mapping of dataset to cutoff
        return {x: y for x, y in zip(['tumor', 'normal', 'gtex'], cutoffs)}

    def gene_kde(self, gene, tissue):
        """
        Returns KDE of gene expression (log2) for given tissue

        :param str gene: Gene (ex: ERBB2) to select
        :param str tissue: Tissue (ex: Breast) to select
        :return: Returns holoviews Overlay object of gene KDE
        :rtype: hv.Overlay
        """
        # Subset dataframe by tissue and gene
        df = self._subset(gene, tissue)

        # Subset by dataset
        tumor, normal, gtex = subset_by_dataset(df)

        # Define x dimension for labeling
        x = hv.Dimension('Gene Expression', unit='log2(x+1)')

        # Create KDE objects
        t = hv.Distribution(tumor[gene].apply(self.l2norm), kdims=[x], label='Tumor-{}'.format(tissue))
        g = hv.Distribution(gtex[gene].apply(self.l2norm), kdims=[x], label='GTEx-{}'.format(tissue))

        return hv.Overlay([t, g], label='{} Expression'.format(gene))

    def multiple_tissue_gene_kde(self, gene, *tissues):
        return hv.Overlay([self.gene_kde(gene, t) for t in tissues],
                          label='{} Expression'.format(gene))

    def gene_distribution(self, gene, tissue_subset=None, extents=None):
        """
        Box and Whisker expression distribution across tissues

        :param str gene: Gene (ex: ERBB2) to select
        :param list tissue_subset: List of tissues to subset by
        :param tuple extents: xmin/ymin/xmax/ymax values
        :return: Returns holoviews BoxWhisker object
        :rtype: hv.BoxWhisker
        """
        # Subset dataframe by gene
        df = self.df[self.df_cols + [gene]].sort_values(gene, ascending=False)

        # Subset by tissues
        if tissue_subset:
            df = df[df.tissue.isin(tissue_subset)]

        # Normalize gene expression
        df[gene] = df[gene].apply(lambda x: np.log2(x + 1))

        # Subset for Tumor and GTEx
        df = df[((df.tumor == 'yes') | (df.dataset == 'gtex'))]

        # return grouped box and whiskers:
        return hv.BoxWhisker((df.tissue, df.dataset, df[gene]), kdims=['tissue', 'dataset'],
                             vdims='gene', label='{} Expression'.format(gene))

    def gene_de(self, gene, extents=None):
        """
        Scatter plot of differential expression across all tissues

        :param str gene: Gene (ex: ERBB2) to select
        :param tuple extents: xmin/ymin/xmax/ymax values
        :return: Scatterplot of values
        :rtype: hv.Scatter
        """
        # Subset dataframe by gene
        df = self.df[self.df_cols + [gene]].sort_values(gene, ascending=False)

        # Subset by dataset
        tumor, normal, gtex = subset_by_dataset(df)

        # For each tissue, calculate L2FC and mean expression
        records = []
        for tissue in sorted(df.tissue.unique()):
            # Calculate mean expression for TCGA tumor and GTEx
            exp = df[(df.tissue == tissue) & ((df.tumor == 'yes') | (df.dataset == 'gtex'))][gene].apply(
                self.l2norm).median()

            # Calculate tumor and normal expression
            t = tumor[tumor.tissue == tissue][gene].median()
            g = gtex[gtex.tissue == tissue][gene].median()

            # Calculate log2 fold change
            l2fc = log2fc(t, g)

            # Store as record
            records.append((exp, l2fc, tissue))

        # Define dimensions of plot
        kdims = ['Expression']
        vdims = ['L2FC', 'Tissue']

        # Create dataframe
        plot = pd.DataFrame.from_records(records, columns=kdims + vdims)

        if extents:
            return hv.Scatter(plot, kdims=kdims, vdims=vdims, extents=extents)
        else:
            return hv.Scatter(plot, kdims=kdims, vdims=vdims)

    def gene_curves(self, gene, tissue):
        """
        Returns set of 3 plots for tissue / gene given a dataframe of metadata and expression values

        :param str gene: Gene (ex: ERBB2) to select
        :param str tissue: Tissue (ex: Breast) to select
        :return: Returns holoviews Layout object containing 3 plots for selected Tisssue / Gene
        :rtype: hv.Layout
        """
        # Subset dataframe for gene and tissue
        df = self._subset(gene, tissue)

        # Logscale gene for calculations
        df[gene] = df[gene].apply(lambda x: np.log2(x + 1))

        # Subset by dataset
        tumor, normal, gtex = subset_by_dataset(df)

        # Get values for plot
        records = []
        for perc_tumor in [x * 0.1 for x in xrange(1, 11)]:
            # Get log2 expression value for top x% tumor samples
            exp = float(tumor.iloc[int(len(tumor) * perc_tumor) - 1][gene])

            # Get percentage of samples in GTEx
            perc_normal = (len(gtex[gtex[gene] > exp]) * 1.0) / len(gtex)

            # Compute L2FC for tumor sample subset vs GTEx
            tumor_mean = tumor.iloc[:int(len(tumor) * perc_tumor) - 1][gene].apply(lambda x: 2 ** x - 1).median()
            gtex_mean = gtex[gene].apply(lambda x: 2 ** x - 1).median()
            l2fc = log2fc(tumor_mean, gtex_mean)

            # Store
            records.append((tissue, exp, l2fc, perc_tumor, perc_normal, len(gtex), len(tumor), 'GTEx'))

        # Create dataframe from records
        info = pd.DataFrame.from_records(records, columns=['tissue', 'expression',
                                                           'l2fc',
                                                           'percent_tumor',
                                                           'percent_normal',
                                                           'num_normals', 'num_tumors',
                                                           'normal_dataset'])

        # Define dimensions
        tissue_dim = hv.Dimension('tissue', label='Tissue')
        ptumor_dim = hv.Dimension('percent_tumor', label='% Tumor')
        pnormal_dim = hv.Dimension('percent_normal', label='percent')
        l2fc_dim = hv.Dimension('l2fc', label='log2FC')
        exp_dim = hv.Dimension('expression', label='log2(x+1)')

        # First plot - Percentage of Normal Samples
        c1 = hv.Curve(data=info, kdims=[ptumor_dim],
                      vdims=[pnormal_dim, tissue_dim], group='Percentage of Normal Samples',
                      extents=(None, 0, None, 1))

        s1 = hv.Scatter(data=info, kdims=[ptumor_dim],
                        vdims=[pnormal_dim, tissue_dim], group='Percentage of Normal Samples')

        # Second Plot - Expression
        c2 = hv.Curve(data=info, kdims=[ptumor_dim],
                      vdims=[exp_dim, tissue_dim], group='Gene Expression',
                      extents=(None, 0, None, 16))

        s2 = hv.Scatter(data=info, kdims=[ptumor_dim],
                        vdims=[exp_dim, tissue_dim], group='Gene Expression')

        # Third Plot - Log2 Fold Change
        c3 = hv.Curve(data=info, kdims=[ptumor_dim],
                      vdims=[l2fc_dim, tissue_dim], group='Log2 Fold Change',
                      extents=(None, -0.5, None, 8))

        s3 = hv.Scatter(data=info, kdims=[ptumor_dim],
                        vdims=[l2fc_dim, tissue_dim], group='Log2 Fold Change')

        return (c1 * s1 + c2 * s2 + c3 * s3).cols(1)

    def sample_counts(self):
        """
        Bargraph of tissues grouped by dataset

        :return:
        """
        pass

    def differential_expression_comparison(self):
        """
        Categorical scatterplot of concordance between tissues for gene differential expression

        :return:
        """
        pass

    def trimap(self, genes, title, tissue_subset=None, num_neighbors=50):
        """
        Dimensionality reduction via Trimap

        :param list(str) genes: List of genes to subset by
        :param list(str) tissue_subset: List of tissues to subset by
        :return: Scatterplot of dimensionality reduction
        :rtype: hv.Scatter
        """
        # Subset dataframe by genes (keeping some metadata)
        df = self.df[self.df_cols + genes]

        # Subset by tissues
        if tissue_subset:
            df = df[df.tissue.isin(tissue_subset)]

        # Run Trimap (used to be called t-ETE)
        z = run_tete(df[genes], num_dims=2, num_neighbors=num_neighbors)

        # Add results to dataframe
        df['x'] = z[:, 0]
        df['y'] = z[:, 1]

        return hv.Scatter(df, kdims=['x'], vdims=['y'] + self.df_cols,
                          group=title)

    def tsne(self, genes, title, tissue_subset=None, perplexity=50, learning_rate=1000):
        """
        Dimensionality reduction via t-SNE

        :param list(str) genes: List of genes to subset by
        :param list(str) tissue_subset: List of tissues to subset by
        :return: Scatterplot of dimensionality reduction
        :rtype: hv.Scatter
        """
        # Subset dataframe by genes (keeping some metadata)
        df = self.df[self.df_cols + genes]

        # Subset by tissues
        if tissue_subset:
            df = df[df.tissue.isin(tissue_subset)]

        # Run t-SNE
        z = run_tsne(df[genes], num_dims=2, perplexity=perplexity, learning_rate=learning_rate)

        # Add results to dataframe
        df['x'] = z[:, 0]
        df['y'] = z[:, 1]

        return hv.Scatter(df, kdims=['x'], vdims=['y'] + self.df_cols,
                          group=title)



import os
import re

import pandas as pd

from rnaseq_lib.data import load_ucsf_genes, load_gene_map, load_samples
from rnaseq_lib.gtf import get_protein_coding_genes
from rnaseq_lib.utils import flatten

__location__ = os.path.realpath(os.path.join(os.getcwd(), os.path.dirname(__file__)))


# dtype for metadata-expression dataframe
dtype = {'id': object,
         'reads': int,
         'size_MB': float,
         'sex': object,
         'tissue': object,
         'seq_site': object,
         'weight': float,
         'height': float,
         'mapped_reads': float,
         'race': object,
         'age': float,
         'qc': float,
         'dataset': object,
         'tumor': object,
         'type': object}


def map_genes(genes, strict=True):
    """
    Maps gene IDs to gene names

    :param list genes: ENSEMBL gene IDs to be mapped to gene names
    :param bool strict: If true, raies a KeyError if gene is not found in the gene_map
    :return: Mapped genes
    :rtype: list
    """
    gene_map = load_gene_map()
    if strict:
        return [gene_map[x.split('.')[0]] for x in genes]
    else:
        mapped = []
        for g in genes:
            try:
                mapped.append(gene_map[g.split('.')[0]])
            except KeyError:
                mapped.append(g)
        return mapped


def get_ucsf_subset(df, strict=False):
    """
    Subset UCSF dataframe and return.

    :param pd.DataFrame df: Input Dataframe in the format of "Genes by Samples"
    :param bool strict: If True, raises an error if gene is unmapped
    :return: Subset of Dataframe that only includes UCSF genes
    :rtype: pd.DataFrame
    """
    df.index = map_genes(df.index, strict=strict)

    ucsf_genes = load_ucsf_genes()
    ucsf_genes = [x for x in ucsf_genes if x in df.index]

    return df.loc[ucsf_genes]


def get_tumor_samples(tissue):
    """
    Returns TCGA tumor samples for a tissue
    :param str tissue: Tissue to grab TCGA tumor samples from
    :return: List of tumor samples
    :rtype: list
    """
    return load_samples()[tissue]['tcga_t']


def get_gtex_samples(tissue):
    """
    Returns GTEx samples for a tissue

    :param str tissue: Tissue to grab GTEx samples from
    :return: List of GTEx samples
    :rtype: list
    """
    return load_samples()[tissue]['gtex']


def get_normal_samples(tissue):
    """
    Returns TCGA normal samples for a tissue

    :param str tissue: Tissue to grab TCGA normal samples from
    :return: List of TCGA normal samples
    :rtype: list
    """
    return load_samples()[tissue]['tcga_n']


def identify_tissue_from_str(content):
    """
    Identifies possible tissue(s) referenced by a given string by matching terms that are "complete words"

    :param str content: Text to examine for terms associated with tissues
    :return: Possible tissues referenced in input string
    :rtype: set(str)
    """
    td_map = tissue_disease_mapping()

    # \b Matches the empty string, but only at the beginning or end of a word
    return set([k for k, v in td_map.iteritems() if
                any([term for term in v if re.search(r'\b{}\b'.format(term), content, flags=re.IGNORECASE)])])


def subset_by_dataset(df):
    """
    Returns tumor, normal, and GTEx dataframe subsets

    :param pd.Dataframe df: Dataframe containing a "dataset" and "tumor" column
    :return: One dataframe per dataset
    :rtype: tuple(pd.Dataframe, pd.Dataframe, pd.Dataframe)
    """
    tumor = df[(df.dataset == 'tcga') & (df.tumor == 'yes')]
    normal = df[(df.dataset == 'tcga') & (df.tumor == 'no')]
    gtex = df[(df.dataset == 'gtex')]
    return tumor, normal, gtex


def tissue_disease_mapping():
    """
    Maps tissue types to words associated with cancers of that tissue

    :return: Tissue / disease term mapping
    :rtype: dict(str, list(str))
    """
    return {
        'Adrenal': ['adrenal', 'adrenocortical', 'cortical', 'oncocytic', 'myxoid'],
        'Bladder': ['bladder', 'urothelial'],
        'Blood': ['blood', 'leukemia', 'myeloma', 'hematologic'],
        'Bone': ['bone', 'osteosarcoma', 'ewing'],
        'Brain': ['brain', 'astrocytoma', 'neurocytoma', 'choroid', 'plexus', 'neuroepithelial',
                  'ependymal', 'Pleomorphic Xanthoastrocytoma', 'fibrillary', 'giant-cell', 'glioblastoma',
                  'multiforme', 'gliomatosis', 'cerebri', 'gliosarcoma',
                  'hemangiopericytoma', 'medulloblastoma', 'medulloepithelioma', 'meningeal', 'neuroblastoma',
                  'neurocytoma', 'oligoastrocytoma', 'optic', 'ependymoma', 'pilocytic', 'pinealoblastoma',
                  'pineocytoma', 'meningioma', 'subependymoma', 'retinoblastoma', 'neuro', 'glioma', 'ganglioglioma',
                  'oligodendroglioma', 'ganglioglioma'],
        'Breast': ['breast'],
        'Cervix': ['cervix', 'cervical'],
        'Colon': ['colon', 'rectal', 'colorectal', 'intestine', 'intestinal', 'bowel', 'lynch', 'rectum'],
        'Esophagus': ['esophagus', 'barrett', 'esophageal', 'oropharyngeal', 'oropharynx', 'laryngeal'],
        'Kidney': ['kidney', 'renal ', 'nephron', 'nephroma', 'wilm', 'chromophobe', 'rhabdoid', 'papillary renal'],
        'Liver': ['liver', 'hepatic', 'Hepatocellular', 'parenchymal', 'cholang'],
        'Lung': ['lung', 'small-cell carcinoma', 'non-small-cell carcinoma', 'small cell carcinoma',
                 'non small cell carcinoma', 'non small-cell carcinoma', 'mesothelioma'],
        'Ovary': ['ovary', 'ovarian', 'endometrioid', 'fallopian', 'cord', 'bronchogenic'],
        'Pancreas': ['pancreas', 'pancreatic', 'cystadenocarcinomas'],
        'Prostate': ['prostate'],
        'Skin-Head': ['head', 'neck', 'skin', 'basal', 'melanoma', 'oral', 'merkel', 'salivary'],
        'Stomach': ['stomach', 'gastric', 'bile', 'cholangiocarcinoma', 'Gastrointestinal'],
        'Testis': ['testis', 'testicular', 'testes', 'gonad', ],
        'Thyroid': ['thyroid', 'papillary carcinoma'],
        'Uterus': ['uterus', 'uterine', 'endometrial', 'ureteral', 'gestational']
    }


def grep_cancer_terms(content, replace_newlines_with_periods=True, comprehensive=False):
    """
    Returns sentences with cancer terms

    :param str content: String containing sentences to check for cancer terms
    :param bool replace_newlines_with_periods: If True, replaces newlines with periods so they count as "sentences"
    :param bool comprehensive: if True, adds all values from tissue_disease_mapping
    :return: Sentences with matches
    :rtype: list(str)
    """
    terms = {'cancer', 'leukemia', 'carcinoma', 'squamous', 'lymphoma', 'malignant',
             'metastasis', 'metastatic', 'sarcoma', 'tumor', 'adenocarcinoma'}

    # Add all terms from tissue_disease_mapping to grep list
    terms = terms.union(set(flatten(tissue_disease_mapping().values()))) if comprehensive else terms

    # Replace newlines with periods
    content = content.replace('\n', '.') if replace_newlines_with_periods else content

    # Return sentences that include terms
    return [x for x in content.split('.') if any(y for y in terms if y.upper() in x.upper())]


def validate_genes(input_genes):
    from rnaseq_lib.web import find_gene_given_alias

    # Check if user submits single gene as a string
    input_genes = [input_genes] if type(input_genes) == str else input_genes

    # Create valid_genes list
    gene_map = load_gene_map()
    valid_genes = set(gene_map.keys() + gene_map.values())

    # Iterate through genes
    genes = []
    for g in input_genes:

        # Check if listed Gene is valid
        gene = None
        if g in valid_genes:
            gene = g

        # If gene invalid, search MyGene
        if not gene:
            gene = find_gene_given_alias(g)

        # Append gene or None
        genes.append(gene)

    return genes


def subset_protein_coding_genes(df, gtf_path):
    """
    Return dataframe with protein-coding genes

    :param pd.DataFrame df: Genes by samples gene expression dataframe
    :param str gtf_path:
    :return: Subset dataframe
    :rtype: pd.DataFrame
    """
    genes = get_protein_coding_genes(gtf_path)
    return df.loc[genes]



import os
from subprocess import call


def resize_image(img_path, resize='28x28', grayscale=True, output_prefix='resized.'):
    """
    Resizes and image using ImageMagick

    :param str img_path: Path to image
    :param str resize: Resize w x h in pixels
    :param bool grayscale: If True, image will be greyscale
    :param str output_prefix: Prefix for output image
    :return: Output path
    :rtype: str
    """
    dirname, name = os.path.split(img_path)

    params = ['convert', img_path,
              '-resize', resize + '^',
              '-gravity', 'center',
              '-extent', resize]

    if grayscale:
        params.extend(['-type', 'Grayscale'])

    output_path = os.path.join(dirname, output_prefix + name)
    call(params + [output_path])

    return output_path



import holoviews as hv

gene_curves_opts = {
    'Curve': {'plot': dict(height=120, width=600, tools=['hover'], invert_xaxis=True, yrotation=45, yaxis='left'),
              'style': dict(line_width=1.5)},
    'Curve.Percentage_of_Normal_Samples': {'plot': dict(xaxis=None, invert_yaxis=True),
                                           'style': dict(color='Blue')},
    'Curve.Gene_Expression': {'plot': dict(xaxis=None),
                              'style': dict(color='Green')},
    'Curve.Log2_Fold_Change': {'plot': dict(height=150),
                               'style': dict(color='Purple')},
    'Scatter': {'style': dict(color='red', size=3)}}

gene_kde_opts = {
    'Distribution': {'plot': dict(filled=False),
                    'style': dict(line_color=hv.Cycle())}
}

gene_distribution_opts = {
    'BoxWhisker': {'plot': dict(width=900, xrotation=45)}
}

gene_de_opts = {
    'Scatter': {'plot': dict(color_index='Tissue', legend_position='left', width=700, height=500, tools=['hover']),
                'style': dict(cmap='tab20', size=10, alpha=0.5)}
}



import os

from synapseclient import Synapse, File

expression = 'syn11311347'
metadata = 'syn11311931'
objects = 'syn11515014'
pathway = 'syn11585314'


def upload_file(file_path, login, parent, description=None):
    """
    Uploads file to Synapse. Password must be stored in environment variable SYNAPSE_PASS

    :param str file_path: Path to file
    :param str login: Login (usually an email address)
    :param str parent: Parent Synapse ID (example: syn12312415) where file will be placed
    :param str description: Optional description to add
    """
    description = '' if None else description
    f = File(file_path, description=description, parent=parent)

    syn = _syn_login(login)
    syn.store(f)


def download_file(synid, login, download_location='.'):
    """
    Synapse ID of file to download

    :param str synid: Synapse ID
    :param str login: Synapse ID
    :param str download_location: Download location for file
    """
    syn = _syn_login(login)
    syn.get(synid, downloadLocation=download_location)


def _syn_login(login):
    """
    Login to synapse. Set environment variable SYNAPSE_PASS to the password for `login`

    :param str login:
    :return: Synapse instance
    :rtype: instance
    """
    assert 'SYNAPSE_PASS' in os.environ, 'SYNAPSE_PASS must be set as an environment variable'
    syn = Synapse()
    syn.login(login, os.environ['SYNAPSE_PASS'])
    return syn
